import underworld as uw
import numpy as np
from .scaling import nonDimensionalize as nd
from .scaling import UnitRegistry as u


class TemperatureBCs(object):

    def __init__(self, Model, left=None, right=None, top=None, bottom=None,
                 front=None, back=None, indexSets=None, materials=None,
                 bottom_material=None, top_material=None, left_material=None,
                 right_material=None, back_material=None, front_material=None):

        self.Model = Model
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom
        self.front = front
        self.back = back
        self.indexSets = indexSets
        self.materials = materials
        self.bottom_material = bottom_material
        self.top_material = top_material
        self.left_material = left_material
        self.right_material = right_material
        self.back_material = back_material
        self.front_material = front_material

    def __getitem__(self, name):
        return self.__dict__[name]

    @staticmethod
    def _get_heat_flux(side, heat_flow, material):

        if not material.capacity:
            raise ValueError("""Material {0} has no capacity
                             defined""".format(material.name))

        if not material.density.reference_density:
            raise ValueError("""Material {0} has no density
                             defined""".format(material.name))

        cp = material.capacity
        rho = material.density.reference_density
        return heat_flow / (rho * cp)

    @staticmethod
    def _check_temp(val):
        if isinstance(val, u.Quantity):
            return (val.units == u.degK) or (val.units == u.degC)
        return False

    def _check_flux(self, heat_flow):
        if isinstance(heat_flow, u.Quantity):
            val = heat_flow.to_base_units()
            return (val.units == u.kilogram * u.second**-3)
        return False

    def get_conditions(self):

        Model = self.Model
        # Reinitialise neumnann and dirichlet condition
        self.dirichlet_indices = []
        self.neumann_indices = []

        self.dirichlet_indices.append(Model.mesh.specialSets["Empty"])
        self.neumann_indices.append(Model.mesh.specialSets["Empty"])

        # A lot of the following code is redundant and could be simplified.
        # It is however rather easy to follow

        self.dirichlet_indices = [Model.mesh.specialSets["Empty"]]
        if self.left is not None:
            if self._check_temp(self.left):
                Model.temperature.data[Model._left_wall.data] = nd(self.left)
                self.dirichlet_indices[0] += Model._left_wall
            elif self._check_flux(self.left):
                if not self.left_material:
                    raise ValueError("left_material is missing")
                material = self.left_material
                self.neumann_indices[0] += Model._left_wall
                heat_flow = self._get_heat_flux("left", self.left, material)
                Model._heatFlux.data[Model._left_wall.data] = (
                    nd(heat_flow))

        if self.right is not None:
            if self._check_temp(self.right):
                Model.temperature.data[Model._right_wall.data] = nd(self.right)
                self.dirichlet_indices[0] += Model._right_wall
            elif self._check_flux(self.right):
                if not self.right_material:
                    raise ValueError("right_material is missing")
                material = self.right_material
                self.neumann_indices[0] += Model._right_wall
                heat_flow = self._get_heat_flux("right", self.right, material)
                Model._heatFlux.data[Model._right_wall.data] = (
                    nd(heat_flow))

        if self.top is not None:
            if self._check_temp(self.top):
                Model.temperature.data[Model._top_wall.data] = nd(self.top)
                self.dirichlet_indices[0] += Model._top_wall
            elif self._check_flux(self.top):
                if not self.top_material:
                    raise ValueError("top_material is missing")
                material = self.top_material
                self.neumann_indices[0] += Model._top_wall
                heat_flow = self._get_heat_flux("top", self.top, material)
                Model._heatFlux.data[Model._top_wall.data] = (
                    nd(heat_flow))

        if self.bottom is not None:
            if self._check_temp(self.bottom):
                Model.temperature.data[Model._bottom_wall.data] = (
                    nd(self.bottom))
                self.dirichlet_indices[0] += Model._bottom_wall
            elif self._check_flux(self.bottom):
                if not self.bottom_material:
                    raise ValueError("bottom_material is missing")
                material = self.bottom_material
                self.neumann_indices[0] += Model._bottom_wall
                heat_flow = self._get_heat_flux(
                    "bottom", self.bottom, material)
                Model._heatFlux.data[Model._bottom_wall.data] = (
                    nd(heat_flow))

        if self.back is not None and Model.mesh.dim > 2:
            if self._check_temp(self.back):
                Model.temperature.data[Model._back_wall.data] = nd(self.back)
                self.dirichlet_indices[0] += Model._back_wall
            elif self._check_flux(self.back):
                if not self.back_material:
                    raise ValueError("back_material is missing")
                material = self.back_material
                self.neumann_indices[0] += Model._back_wall
                heat_flow = self._get_heat_flux("back", self.back, material)
                Model._heatFlux.data[Model._back_wall.data] = (
                    nd(heat_flow))

        if self.front is not None and Model.mesh.dim > 2:
            if self._check_temp(self.front):
                Model.temperature.data[Model._front_wall.data] = nd(self.front)
                self.dirichlet_indices[0] += Model._front_wall
            elif self._check_flux(self.front):
                if not self.front_material:
                    raise ValueError("front_material is missing")
                material = self.front_material
                self.neumann_indices[0] += Model._front_wall
                heat_flow = self._get_heat_flux("front", self.front, material)
                Model._heatFlux.data[Model._front_wall.data] = (
                    nd(heat_flow))

        if self.indexSets:
            for elem in self.indexSets:
                indexSet, temp = elem
                isThere = np.in1d(Model.mesh.data_nodegId, list(indexSet))
                if isThere.any():
                    local_indices = np.arange(Model.mesh.nodesDomain)[isThere]
                    indexSet = uw.mesh.FeMesh_IndexSet(
                        self.Model.mesh, topologicalIndex=0,
                        size=self.Model.mesh.nodesGlobal,
                        fromObject=local_indices)
                    self._check_temp(temp)
                    Model.temperature.data[local_indices] = nd(temp)
                    self.dirichlet_indices[0] += indexSet

        if self.materials:
            for (material, temp) in self.materials:
                if material and nd(temp):
                    self._check_temp(temp)
                    indexSet = Model._get_material_indices(material)
                    Model.temperature.data[indexSet.data] = nd(temp)
                    self.dirichlet_indices[0] += indexSet

        conditions = []

        conditions.append(uw.conditions.DirichletCondition(
            variable=Model.temperature,
            indexSetsPerDof=self.dirichlet_indices))

        neumann_indices = []
        for val in self.neumann_indices:
            if val.data.size > 0:
                neumann_indices.append(val)
            else:
                neumann_indices.append(None)
        neumann_indices = tuple(neumann_indices)

        if neumann_indices != tuple([None]):
            conditions.append(uw.conditions.NeumannCondition(
                fn_flux=Model._heatFlux,
                variable=Model.temperature,
                indexSetsPerDof=self.neumann_indices))

        if not conditions:
            raise ValueError("Undefined conditions")

        return conditions




# Configuration file
from scaling import u
from _validate import *


rcParams = {

    "model.name": ["Model", validate_string],
    "output.directory": ["outputs", validate_path],
    "element.type" : ["Q1/dQ0", validate_string],

    "minimum.viscosity": [1e19 * u.pascal * u.second, validate_quantity],
    "maximum.viscosity": [1e25 * u.pascal * u.second, validate_quantity],

    "swarm.variables" : [["materialField",
                          "plasticStrain",
                          "viscosityField",
                          "timeField",
                          "densityField"], validate_stringlist],
    "mesh.variables" :  [["velocityField",
                          "temperature",
                          "pressureField",
                          "strainRateField",
                          "boundariesField",
                          "projMaterialField",
                          "projTimeField",
                          "projStressField",
                          "projStressTensor",
                          "projViscosityField",
                          "projPlasticStrain",
                          "projDensityField"], validate_stringlist],
    "default.outputs" : [["materialField",
                          "temperature",
                          "pressureField",
                          "densityField",
                          "plasticStrain",
                          "strainRateField",
                          "velocityField",
                          "viscosityField",
                          "timeField",
                          "projStressField",
                          "projTimeField",
                          "projMaterialField",
                          "projViscosityField",
                          "projPlasticStrain",
                          "projDensityField"], validate_stringlist],
    "glucifer.outputs" : [["materialField",
                           "temperature",
                           "viscosityField",
                           "pressureField",
                           "plasticStrain",
                           "velocityField"], validate_stringlist],
    "restart.fields" : [["materialField",
                          "temperature",
                          "pressureField",
                          "plasticStrain",
                          "velocityField"], validate_stringlist],

    "gravity": [9.81 * u.meter / u.second**2, validate_quantity],
    "swarm.particles.per.cell.2D": [40, validate_int],
    "swarm.particles.per.cell.3D": [120, validate_int],

    "popcontrol.aggressive" : [True, validate_bool],
    "popcontrol.split.threshold" : [0.15, validate_float],
    "popcontrol.max.splits" : [10, validate_int],
    "popcontrol.particles.per.cell.2D" : [20, validate_int],
    "popcontrol.particles.per.cell.3D" : [60, validate_int],

    "CFL": [0.5, validate_float],

    "solver" : ["mg", validate_solver],
    "penalty" : [0.0, validate_float],
    "initial.nonlinear.tolerance": [1e-2, validate_float],
    "nonlinear.tolerance": [1e-2, validate_float],
    "maximum.timestep" : [200000, validate_int],
    "initial.nonlinear.min.iterations": [2, validate_int],
    "initial.nonlinear.max.iterations": [500, validate_int],
    "nonlinear.min.iterations": [2, validate_int],
    "nonlinear.max.iterations": [500, validate_int],
    "nonlinear.tolerance.adjust.factor": [2, validate_int],
    "nonlinear.tolerance.adjust.nsteps": [100, validate_int],
    "mg.levels": [None, validate_int_or_none],

    "rheology.default.uppercrust": ["Patterson et al., 1990", validate_viscosity],
    "rheology.default.midcrust": ["Patterson et al., 1990", validate_viscosity],
    "rheology.default.lowercrust": ["Wang et al., 2012", validate_viscosity],
    "rheology.default.mantlelithosphere": ["Hirth et al., 2003", validate_viscosity],
    "rheology.default.mantle": ["Karato and Wu, 1990", validate_viscosity],

    "scaling.length": [1.0 * u.meter, validate_quantity],
    "scaling.mass": [1.0 * u.kilogram, validate_quantity],
    "scaling.time": [1.0 * u.year, validate_quantity],
    "scaling.temperature": [1.0 * u.degK, validate_quantity],
    "scaling.substance": [1.0 * u.mole, validate_quantity],

    "time.SIunits": [u.years, validate_quantity],
    "viscosityField.SIunits" : [u.pascal * u.second, validate_quantity],
    "densityField.SIunits" : [u.kilogram / u.metre**3, validate_quantity],
    "velocityField.SIunits" : [u.centimeter / u.year, validate_quantity],
    "temperature.SIunits" : [u.degK, validate_quantity],
    "pressureField.SIunits" : [u.pascal , validate_quantity],
    "projStressTensor.SIunits" : [u.pascal , validate_quantity],
    "projStressField.SIunits" :  [u.pascal , validate_quantity],
    "strainRateField.SIunits" : [1.0 / u.second, validate_quantity],
    "projViscosityField.SIunits"  : [u.pascal * u.second, validate_quantity],
    "projDensityField.SIunits" : [u.kilogram / u.metre**3, validate_quantity],
    "projTimeField.SIunits" : [u.megayears, validate_quantity],

    "useEquationResidual" : [False, validate_bool],
    "alpha": [0., validate_float],
    "shearHeating": [False, validate_bool],
    "surface.pressure.normalization": [True, validate_bool],
    "pressure.smoothing": [True, validate_bool],
    "advection.diffusion.method": ["SUPG", validate_string]
    }





# THIS FILE IS GENERATED FROM UWGeodynamics SETUP.PY
#
underworld_version = 'NOT INSTALLED'
short_version = '0.9.2'
version = '0.9.2'
full_version = '0.9.2-dev-10ac92e(master)'
git_revision = '10ac92ee87a6139b5a1e4ed5f11be23651103025'
release = False

if not release:
    version = full_version



import json
import inspect
from collections import OrderedDict
import UWGeodynamics as GEO
from ._frictional_boundary import FrictionBoundaries
from ._thermal_boundaries import TemperatureBCs
from ._velocity_boundaries import VelocityBCs
from ._visugrid import Visugrid
from json import JSONDecoder


class ObjectEncoder(json.JSONEncoder):

    def default(self, obj):

        if isinstance(obj, GEO.Model):

            model = OrderedDict()

            model["Type"] = "Model"
            model["name"] = obj["name"]

            # Save rcparams (use file)
            with open(GEO.uwgeodynamics_fname(), "r") as f:
                rcParams = f.read()

            model["rcParams"] = rcParams

            model["elementRes"] = obj["elementRes"]
            model["minCoord"] = obj["minCoord"]
            model["maxCoord"] = obj["maxCoord"]
            model["gravity"] = obj["gravity"]
            model["periodic"] = obj["periodic"]
            model["elementType"] = obj["elementType"]

            # Encode Scaling
            scaling = {}
            for key, val in GEO.scaling_coefficients.iteritems():
                scaling[key] = val
            model["scaling"] = scaling

            # Encode velocity boundary conditions
            model["velocityBCs"] = obj.velocityBCs

            # Encode temperature boundary conditions
            model["temperatureBCs"] = obj.temperatureBCs

            model["materials"] = []
            # Encode materials
            for material in reversed(obj.materials):
                if material is obj:
                    #val = super(Model, obj).to_json()
                    #model["materials"].append(val)
                    continue
                else:
                    model["materials"].append(material)

            return model

        # Material
        if isinstance(obj, GEO.Material):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        # Constant Density
        if isinstance(obj, GEO.ConstantDensity):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        # Linear Density
        if isinstance(obj, GEO.LinearDensity):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        # Boundary conditions
        # Frictional Boundary
        if isinstance(obj, FrictionBoundaries):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        # Solidus, Liquidus
        if isinstance(obj, (GEO.Solidus, GEO.Liquidus)):
            d = OrderedDict()
            return d

        # Constant Viscosity
        if isinstance(obj, GEO.ConstantViscosity):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        # Drucker Prager
        if isinstance(obj, GEO.DruckerPrager):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        # Von Mises
        if isinstance(obj, GEO.VonMises):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        # Viscous Creep
        if isinstance(obj, GEO.ViscousCreep):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, TemperatureBCs):
            d = OrderedDict()
            d["Type"] = "Temperature Boundary Conditions"

            attributes = [
                "left",
                "right",
                "top",
                "bottom",
                "back",
                "front",
                "indexSets"]

            for attribute in attributes:
                d[attribute] = obj[attribute]

            return d

        if isinstance(obj, GEO.Balanced_InflowOutflow):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, VelocityBCs):
            d = OrderedDict()
            d["Type"] = "Velocity Boundary Conditions"

            attributes = [
                "left",
                "right",
                "top",
                "bottom",
                "back",
                "front",
                "indexSets"]

            for attribute in attributes:
                d[attribute] = obj[attribute]

            return d

        if isinstance(obj, Visugrid):
            d = OrderedDict()
            d["Type"] = "Visugrid"
            d["minCoord"] = obj.minCoord
            d["maxCoord"] = obj.maxCoord
            d["elementRes"] = obj.elementRes
            return d

        # Shapes
        if isinstance(obj, GEO.shapes.Polygon):
            d = OrderedDict()
            d["Type"] = "Polygon"
            d["vertices"] = obj.vertices
            return d

        if isinstance(obj, GEO.shapes.MultiShape):
            d = OrderedDict()
            d["Type"] = "Multishape"
            d["shapes"] = obj.shapes
            return d

        if isinstance(obj, GEO.shapes.Layer):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO.shapes.Box):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO.shapes.Disk):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO.shapes.Annulus):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO.surfaceProcesses.Badlands):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO.surfaceProcesses.ErosionThreshold):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO.surfaceProcesses.SedimentationThreshold):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO.surfaceProcesses.ErosionAndSedimentationThreshold):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO.LecodeIsostasy):
            args, _, _, _ = inspect.getargspec(obj.__init__)
            d = OrderedDict()
            d["Type"] = type(obj).__name__
            for arg in args[1:]:
                try:
                    d[arg] = getattr(obj, arg)
                except AttributeError:
                    d[arg] = getattr(obj, "_" + arg)
            return d

        if isinstance(obj, GEO._mesh_advector._mesh_advector):
            d = OrderedDict()
            d["Type"] = "_mesh_advector"
            d["axis"] = obj.axis
            return d

        if isinstance(obj, GEO.u.Quantity):
            d = OrderedDict()
            d["value"] = obj.magnitude
            d["units"] = obj.units
            return d

        if isinstance(obj, GEO.u.Unit):
            return str(obj)

        return super(ObjectEncoder, self).default(obj)


class ObjectDecoder(json.JSONDecoder):

    def __init__(self, *args, **kwargs):
        JSONDecoder.__init__(self, object_hook=self.json_to_model,
                             *args, **kwargs)

    def json_to_model(self, d):

        if "Type" in d and d["Type"] == "Model":
            model = GEO.Model
            args, _, _, _ = inspect.getargspec(model.__init__)
            kwargs = {}
            for arg in args:
                if arg != "self":
                    kwargs[arg] = self.json_to_model(d[arg])

            return model(**kwargs)

        if "units" in d:
            value = d["value"]
            units = d["units"]
            return GEO.u.Quantity(value, units)







##~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~##
##                                                                                   ##
##  This file forms part of the Underworld geophysics modelling application.         ##
##                                                                                   ##
##  For full license and copyright information, please refer to the LICENSE.md file  ##
##  located at the project root, or contact the authors.                             ##
##                                                                                   ##
##~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~##
"""
This module implements some basic functionality for posting usage metrics
to google analytics.
"""

import json, httplib, urllib
import uuid
import underworld as uw

GA_TRACKING_ID = "UA-115593169-1"
GA_CLIENT_ID = uw._id

def PostGAEvent( category, action, label=None, value=None ):
    """
    Posts an Event Tracking message to Google Analytics.

    Full parameter reference may be found here:
    https://developers.google.com/analytics/devguides/collection/protocol/v1/parameters#ev

    Note, this function will return quietly on any errors.

    Parameters
    ----------
    category: str
        Textual name for event category.
    action: str
        Textual name for event action.
    label: str
        Optional label for event.
    value: non-negative integer
        Optional value for event.


    """
    try:
        connection = httplib.HTTPSConnection('www.google-analytics.com')
        form_fields = {
        "v"  : "1",             # Version.
        "aip": "1",             # Enable IP anonymizing.
        "tid": GA_TRACKING_ID,  # Tracking ID / Web property / Property ID.
        "ds" : "app",           # Data Source.
        "cid": GA_CLIENT_ID,    # Anonymous Client ID.
        "t"  : "event",         # Event hit type.
        "an" : "UWGeodynamics",   # Application name.
        "av" : uw.__version__,  # Application version.
        "ec" : category,        # Event Category. Required.
        "ea" : action,          # Event Action. Required.
        "el" : label,           # Event label.
        "ev" : value,           # Event value.
        "cm2": uw.nProcs(),     # Number of processes used. Stored into custom metric 2.
        "cd5": str(uw.nProcs()),# Number of processes used. Stored into custom dim 2. Not sure if necessary.
        }
        import os
        # add user id if set
        if "UW_USER_ID" in os.environ:
            form_fields["uid"] = os.environ["UW_USER_ID"]
            form_fields["cd4"] = os.environ["UW_USER_ID"]

        if "UW_MACHINE" in os.environ:
            form_fields["cd6"] = os.environ["UW_MACHINE"]

        params = urllib.urlencode(form_fields)
        connection.connect()
        connection.request('POST', '/collect?%s' % params, '', { "Content-Type": "application/x-www-form-urlencoded" })
    except:
        pass




    def load( self, filename ):
        """
        Load the swarm variable from disk. This must be called *after* the swarm.load().

        Parameters
        ----------
        filename : str
            The filename for the saved file. Relative or absolute paths may be
            used, but all directories must exist.

        Notes
        -----
        This method must be called collectively by all processes.


        Example
        -------
        Refer to example provided for 'save' method.

        """

        if not isinstance(filename, str):
            raise TypeError("'filename' parameter must be of type 'str'")

        if self.swarm._checkpointMapsToState != self.swarm.stateId:
            raise RuntimeError("'Swarm' associate with this 'SwarmVariable' does not appear to be in the correct state.\n" \
                               "Please ensure that you have loaded the swarm prior to loading any swarm variables.")
        gIds = self.swarm._local2globalMap

        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        nProcs = comm.Get_size()

        # open hdf5 file
        h5f = h5py.File(name=filename, mode="r", driver='mpio', comm=MPI.COMM_WORLD)

        dset = h5f.get('data')
        if dset == None:
            raise RuntimeError("Can't find 'data' in file '{}'.\n".format(filename))

        if dset.shape[1] != self.data.shape[1]:
            raise RuntimeError("Cannot load file data on current swarm. Data in file '{0}', " \
                               "has {1} components -the particlesCoords has {2} components".format(filename, dset.shape[1], self.particleCoordinates.data.shape[1]))

        particleGobalCount = self.swarm.particleGlobalCount
        
        if dset.shape[0] != particleGobalCount:
            if rank == 0:
                import warnings
                warnings.warn("Warning, it appears {} particles were loaded, but this h5 variable has {} data points. Perhaps you restarting on a new mesh geometry?". format(particleGobalCount, dset.shape[0]), RuntimeWarning)

        # for efficiency, we want to load swarmvariable data in the largest stride chunks possible.
        # we need to determine where required data is contiguous.
        # first construct an array of gradients. the required data is contiguous
        # where the indices into the array are increasing by 1, ie have a gradient of 1.
        gradIds = np.zeros_like(gIds)            # creates array of zeros of same size & type
        if len(gIds) > 1:
            gradIds[:-1] = gIds[1:] - gIds[:-1]  # forward difference type gradient

        guy = 0
        while guy < len(gIds):

            # do contiguous
            start_guy = guy
            while gradIds[guy]==1:  # count run of contiguous. note bounds check not required as last element of gradIds is always zero.
                guy += 1
            # copy contiguous chunk if found.. note that we are copying 'plus 1' items
            if guy > start_guy:
                self.data[start_guy:guy+1] = dset[gIds[start_guy]:gIds[guy]+1]
                guy += 1

            # do non-contiguous
            start_guy = guy
            while guy<len(gIds) and gradIds[guy]!=1:  # count run of non-contiguous
                guy += 1
            # copy non-contiguous items (if found) using index array slice
            if guy > start_guy:
                self.data[start_guy:guy,:] = dset[gIds[start_guy:guy],:]
                
            # repeat process until all done

        h5f.close();



import glucifer
import underworld as uw
from mpi4py import MPI
from .scaling import nonDimensionalize as nd
from .scaling import Dimensionalize
from .scaling import UnitRegistry as u
from . import rcParams
from .Underworld_extended import SwarmVariable
from scipy import interpolate
import numpy as np

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()


def _visugrid_drawing_object(Model, visugrid):

    m_minx, m_maxx = _get_minmax_coordinates_mesh(Model.mesh, 0)
    m_miny, m_maxy = _get_minmax_coordinates_mesh(Model.mesh, 1)
    model_domain = np.array([(m_minx, m_miny), (m_maxx, m_maxy)])

    v_minx, v_maxx = _get_minmax_coordinates_mesh(visugrid.mesh, 0)
    v_miny, v_maxy = _get_minmax_coordinates_mesh(visugrid.mesh, 1)
    # visugrid_domain = np.array([(v_minx, v_miny), (v_maxx, v_maxy)])

    minx = min(m_minx, v_minx)
    miny = min(m_miny, v_miny)
    maxx = max(m_maxx, v_maxx)
    maxy = max(m_maxy, v_maxy)

    full_domain = np.array([(minx, miny), (maxx, maxy)])
    bounds = full_domain[1] - full_domain[0]

    minx = (model_domain[0][0] - full_domain[0][0]) / bounds[0]
    maxx = 1 + (model_domain[1][0] - full_domain[1][0]) / bounds[0]
    miny = (model_domain[0][1] - full_domain[0][1]) / bounds[1]
    maxy = 1 + (model_domain[1][1] - full_domain[1][1]) / bounds[1]

    return (minx[0], maxx[0]), (miny[0], maxy[0])


def _get_minmax_coordinates_mesh(mesh, axis=0):
    """ Return the minimum and maximum coordinates along axis

    parameter:
    ----------
        axis:
            axis

    returns:
    -------
        tuple: minV, maxV

    """
    maxVal = np.zeros((1))
    minVal = np.zeros((1))
    maxVal[0] = mesh.data[:, axis].max()
    minVal[0] = mesh.data[:, axis].min()

    uw.barrier()
    comm.Allreduce(MPI.IN_PLACE, maxVal, op=MPI.MAX)
    comm.Allreduce(MPI.IN_PLACE, minVal, op=MPI.MIN)
    uw.barrier()

    return minVal, maxVal


class Plots(object):

    def __init__(self, Model):

        self.Model = Model

    @property
    def _boundingBox(self):
        minCoord = tuple([nd(val) for val in self.Model.minCoord])
        maxCoord = tuple([nd(val) for val in self.Model.maxCoord])
        boundingBox = (minCoord, maxCoord)
        return boundingBox

    def materialField(self, figsize=None, title="Material Field",
                      colours=None, script=None, cullface=False,
                      mask=None, visugrid=None, onMesh=False,
                      show=True, store=None, quality=3,
                      **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title
        Fig["boundingBox"] = self._boundingBox

        if onMesh:
            Fig.Surface(self.Model.mesh, self.Model.projMaterialField,
                        **kwargs)
        else:
            pts = Fig.Points(self.Model.swarm,
                             fn_colour=self.Model.materialField,
                             colours=colours,
                             cullface=cullface, name=Fig["title"], **kwargs)
            # pts.colourBar["binlabels"] = True
            # pts.colourBar["align"] = "right"
            # pts.colourBar["vertical"] = True
            pts.colourBar.colourMap["discrete"] = True

        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            # Fig.viewer().window()
            Fig.show()

        return Fig

    material = materialField

    def viscosityField(self, figsize=None, title="Viscosity Field",
                       units=u.pascal * u.second, logScale=True,
                       projected=False, cullface=False,
                       script=None, show=True, quality=3,
                       store=None, visugrid=None, **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title + " " + str(units)
        Fig["boundingBox"] = self._boundingBox

        fact = Dimensionalize(1.0, units).magnitude
        Fig.Points(self.Model.swarm, self.Model._viscosityFn * fact,
                   logScale=logScale, name=Fig["title"], cullface=cullface, **kwargs)

        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    viscosity = viscosityField

    def strainRate(self, figsize=None, title="Strain Rate Field",
                   units=1.0 / u.second,
                   cullface=False,
                   logScale=True, colours="coolwarm",
                   script=None, show=True, quality=3,
                   store=None, visugrid=None, **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title + " " + str(units)
        Fig["boundingBox"] = self._boundingBox

        fact = Dimensionalize(1.0, units).magnitude
        Fig.Surface(self.Model.mesh.subMesh,
                    self.Model.strainRateField * fact,
                    cullface=cullface, logScale=logScale, name=Fig["title"],
                    colours=colours,
                    **kwargs)

        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    def density(self, figsize=None, title="Density Field",
                units=u.kilogram / u.metre**3, quality=3,
                script=None, cullface=False, show=True,
                store=None, visugrid=None, **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title + " " + str(units)
        Fig["boundingBox"] = self._boundingBox

        fact = Dimensionalize(1.0, units).magnitude
        Fig.Points(self.Model.swarm,
                   fn_colour=self.Model._densityFn * fact,
                   cullface=cullface, name=Fig["title"], **kwargs)

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    def temperature(self, figsize=None, title="Temperature Field",
                    units=u.degK, script=None, cullface=False,
                    colours="coolwarm", show=True, quality=3,
                    store=None, visugrid=None, **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title + " " + str(units)
        Fig["boundingBox"] = self._boundingBox

        fact = Dimensionalize(1.0, units).magnitude
        Fig.Surface(self.Model.mesh, self.Model.temperature * fact,
                    colours=colours, cullface=cullface, name=Fig["title"],
                    **kwargs)
        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    def pressureField(self, figsize=None, title="Pressure Field",
                      units=u.pascal, cullface=False,
                      script=None, show=True, quality=3,
                      store=None, visugrid=None, **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title + " " + str(units)
        Fig["boundingBox"] = self._boundingBox

        fact = Dimensionalize(1.0, units).magnitude
        Fig.Surface(self.Model.mesh, self.Model.pressureField * fact,
                    cullface=cullface, name=Fig["title"], **kwargs)
        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    def stressField(self, figsize=None, title="Stress Field",
                    units=u.pascal, cullface=False,
                    script=None, show=True, quality=3,
                    store=None, visugrid=None, **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title + " " + str(units)
        Fig["boundingBox"] = self._boundingBox

        fact = Dimensionalize(1.0, units).magnitude
        Fig.Surface(self.Model.mesh, self.Model.projStressField * fact,
                    cullface=cullface, name=Fig["title"], **kwargs)
        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    def velocityField(self, figsize=None, title="Velocity Field",
                      units=u.centimeter / u.year, cullface=False,
                      script=None, show=True, quality=3,
                      store=None, colours='diverge', visugrid=None,
                      arrowlength=0.0,
                      arrownorm=1.0,
                      arrowres=(16,16,16), **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title + " " + str(units)
        Fig["boundingBox"] = self._boundingBox

        fact = Dimensionalize(1.0, units).magnitude
        velmagfield = uw.function.math.sqrt(
            uw.function.math.dot(self.Model.velocityField,
                                 self.Model.velocityField))
        Fig.Surface(self.Model.mesh, velmagfield * fact,colours=colours,
                   cullface=cullface, name=Fig["title"], **kwargs)
        Fig.VectorArrows(self.Model.mesh,
                         self.Model.velocityField,
                         length=arrowlength,
                         normalise=arrownorm,
                         resolution=arrowres,
                         **kwargs)
        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    def plasticStrain(self, figsize=None, title="Plastic Strain",
                      cullface=False, script=None, show=True,
                      store=None, visugrid=None, quality=3,
                      **kwargs):

        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title
        Fig["boundingBox"] = self._boundingBox

        Fig.Points(self.Model.swarm, fn_colour=self.Model.plasticStrain,
                   cullface=cullface, name=Fig["title"], **kwargs)
        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)
        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    plastic_strain = plasticStrain

    def melt_fraction(self, figsize=None, title="Melt fraction",
                      cullface=False, script=None, show=True,
                      store=None, visugrid=None, quality=3,
                      **kwargs):

        self.Model.update_melt_fraction()
        Fig = glucifer.Figure(store=store, figsize=figsize,
                              quality=quality,
                              min=self._boundingBox[0],
                              max=self._boundingBox[1])
        Fig["title"] = title
        Fig["boundingBox"] = self._boundingBox

        Fig.Points(self.Model.swarm, fn_colour=self.Model.meltField,
                   cullface=cullface, name=Fig["title"], **kwargs)

        if visugrid:
            clip_X, clip_Y = _visugrid_drawing_object(self.Model, visugrid)
            Fig.Mesh(visugrid.mesh, xmin=clip_X[0], xmax=clip_X[1],
                     ymin=clip_Y[0], ymax=clip_Y[1])

        Fig.script(script)

        if show and glucifer.lavavu.is_notebook():
            Fig.show()

        return Fig

    def profile(self,
                start_point,
                end_point,
                field,
                field_units=u.dimensionless,
                distance_units=u.dimensionless,
                npoints=1000):

        if (field not in rcParams["mesh.variables"] and
           field not in rcParams["swarm.variables"]):
            raise ValueError("""{0} is not a valid field, \n
                                Valid fields are {1} or {2}""".format(
                             field, rcParams["mesh.variables"],
                             rcParams["swarm.variables"]))

        start_point = np.array([nd(val) for val in start_point])
        end_point = np.array([nd(val) for val in end_point])

        points = [start_point, end_point]
        x_coords, y_coords = zip(*points)

        x = np.linspace(x_coords[0], x_coords[1], npoints)

        if x_coords[0] == x_coords[1]:
            y = np.linspace(y_coords[0], y_coords[1], npoints)
        else:
            f = interpolate.interp1d(x_coords, y_coords)
            y = f(x)

        dx = np.diff(x)
        dy = np.diff(y)
        distances = np.zeros(x.shape)
        distances[1:] = np.sqrt(dx**2 + dy**2)
        distances = np.cumsum(distances)

        obj = getattr(self.Model, field)

        if isinstance(obj, SwarmVariable):
            obj = getattr(self.Model, "proj" + field[0].upper() + field[1:])

        pts = np.array(zip(x, y))

        values = obj.evaluate(pts)

        distance_fact = Dimensionalize(1.0, distance_units)
        field_fact = Dimensionalize(1.0, field_units)

        distances *= distance_fact.magnitude
        values *= field_fact.magnitude

        return distances, values

    def christmas(self, position):
        return self.profile(start_point=[position, self.Model.top],
                            end_point=[position, self.Model.bottom],
                            field="viscosityField")






# List of Rheology setups
Parameters = {}

Parameters["Resolution"] = [(256,128),
                            (512,256),
                            (1024,512),
                            (2048,1024)]

Parameters["Average Method"] = ["arithmetic",
                                "geometric",
                                "harmonic",
                                "maximum",
                                "minimum",
                                "root mean square"]
